//
//  mbedtls_ec_standalone_test.c
//
//  Created by Persimmon on 8/30/21.
//  Copyright (c) Persimmon
// 
//
// Permission is granted to Nordic Semicondactor Inc. to use this code for the purpuse of
// finding the bug in this software.

#include <mbedtls/bignum.h>
#include <mbedtls/platform.h>
#include <mbedtls/ecdsa.h>
#include <mbedtls/entropy.h>
#include <mbedtls/hmac_drbg.h>
#include <mbedtls/error.h>

#define ok_or_cleanup(rc) if(rc!=0) goto cleanup;

#if 1
    #include <logging/log.h>
    LOG_MODULE_DECLARE(app);
    #define mbedtls_printf LOG_INF
#endif

typedef struct ec_t{
    int (*f_rng)(void *, unsigned char *, size_t);
    void *p_rng;
    mbedtls_entropy_context entropy;
    mbedtls_hmac_drbg_context hmac_drbg;
    struct mbedtls_ecp_group grp;
}ec_t;

static int ec_init(ec_t *ec);
static void ec_free(ec_t *ec);
static void print_err(int error);
static int ec_sign_verify(ec_t *ec);

int mbedtls_ec_standalone_test(void){
    //static ec_t ec; //<---------------------- This doesn't generate a fault
    ec_t ec;

    int rc = ec_init(&ec); ok_or_cleanup(rc);
    rc = ec_sign_verify(&ec); if(rc!=0) print_err(rc);

cleanup:
    ec_free(&ec);
    return rc;
}

static int ec_init(ec_t *ec)
{
#ifdef MBEDTLS_THREADING_C
    mbedtls_printf("&(ec->entropy.mutex) = %p", &(ec->entropy.mutex));
#endif

    mbedtls_entropy_init(&ec->entropy);
    mbedtls_ecp_group_init(&ec->grp);
    mbedtls_hmac_drbg_init(&ec->hmac_drbg);

    const mbedtls_md_info_t *md_info = mbedtls_md_info_from_type(MBEDTLS_MD_SHA256);
    mbedtls_hmac_drbg_set_prediction_resistance(&ec->hmac_drbg, MBEDTLS_HMAC_DRBG_PR_ON);

    //Set the initial seed and set up the entropy source for future reseeds.
    const unsigned char *personalization = (const unsigned char *)  "c5460c341157f0c84c6615ba78247d617369c9304a249bbe59b2cba0cd86118e";
    int rc = mbedtls_hmac_drbg_seed(&ec->hmac_drbg, md_info, mbedtls_entropy_func,
                                    &ec->entropy, personalization,
                                    sizeof(personalization));
    ok_or_cleanup(rc)

    ec->f_rng   = mbedtls_hmac_drbg_random;
    ec->p_rng = &ec->hmac_drbg;

    rc = mbedtls_ecp_group_load(&ec->grp, MBEDTLS_ECP_DP_SECP256K1);
    //rc = mbedtls_ecp_group_load(&ec->grp, MBEDTLS_ECP_DP_SECP256R1);

cleanup:
    return rc;
}

static void ec_free(ec_t *ec)
{
    if (ec == NULL) return;
    mbedtls_hmac_drbg_free(&ec->hmac_drbg);
    mbedtls_entropy_free(&ec->entropy);
    mbedtls_ecp_group_free(&ec->grp);
}

static void print_err(int error)
{
    if(error==0) return;

    char *buf = mbedtls_calloc(300, 1);
    if (buf==NULL){
        mbedtls_printf("\n cannot allocate buffer to print error message\n");
        return;
    }

    mbedtls_strerror(error, buf, 300);
    mbedtls_printf("error: %s\n", buf);
    mbedtls_free(buf);
}

static int ec_sign_verify(ec_t *ec){
    //- init mbedtls first
    mbedtls_mpi r, s, priv;
    mbedtls_ecp_point pub;

    mbedtls_ecp_point_init(&pub);
    mbedtls_mpi_init(&r);
    mbedtls_mpi_init(&s);
    mbedtls_mpi_init(&priv);

    // - generate keys
    int rc = mbedtls_ecp_gen_privkey(&ec->grp, &priv, ec->f_rng, ec->p_rng); ok_or_cleanup(rc)
    rc = mbedtls_ecp_mul(&ec->grp, &pub, &priv, &ec->grp.G, ec->f_rng, ec->p_rng); ok_or_cleanup(rc)


    unsigned char *msg = (unsigned char *)"01234567890123456789012345678901";
    rc = mbedtls_ecdsa_sign(&ec->grp, &r, &s, &priv, msg, 32, ec->f_rng, ec->p_rng);
    if (rc != 0) {
        mbedtls_printf("**: mbedtls_ecdsa_sign() failed. %x\n", -rc);
    }
    ok_or_cleanup(rc)
    rc = mbedtls_ecdsa_verify(&ec->grp, msg, 32, &pub, &r, &s);
    if (rc != 0) {
        mbedtls_printf("**: mbedtls_ecdsa_verify() failed. %x\n", -rc);
    }

cleanup:
    mbedtls_ecp_point_free(&pub);
    mbedtls_mpi_free(&r);
    mbedtls_mpi_free(&s);
    mbedtls_mpi_free(&priv);

    return(rc);
}
